@@ -1,7 +1,7 @@
 #!/usr/bin/perl
 
 package Directory::Transactional::TXN::Root;
-our $VERSION = '0.02';
+our $VERSION = '0.03';
 
 use Squirrel;
 
@@ -18,6 +18,11 @@ has global_lock => (
 	is  => "ro",
 );
 
+# used for auto commit
+has auto_handle => (
+	is  => "ro",
+);
+
 has backup => (
 	isa => "Str",
 	is  => "ro",
@@ -42,6 +47,15 @@ sub all_changed {
 	shift->changed;
 }
 
+sub DEMOLISH {
+	my $self = shift;
+
+	if ( my $ah = $self->auto_handle ) {
+		# in case of rollback
+		$ah->finished(1);
+	}
+}
+
 __PACKAGE__->meta->make_immutable;
 
 __PACKAGE__
